"""
Various utility classes and functions to be used with the taylorism package.
"""

import multiprocessing
from multiprocessing import sharedctypes


class SharedNumpyArray:
    """
    Wrapper to multiprocessing.Array, for it to be shared in memory among
    Workers, while being handled as a numpy.ndarray or numpy.ma.masked_array.
    """

    def __init__(self, array):
        """

        :param array: initialize the SharedNumpyArray with this one, supposed
            to be either a multiprocessing.Array, a numpy.ndarray or a
            numpy.ma.masked_array.

        * constructing a SharedNumpyArray from a multiprocessing.Array will
          not duplicate data but make the SharedArray be a pointer to the
          initial array;
        * constructing from a numpy array will duplicate data, and so the
          initial array will no longer be consistent with this one.

        The process-safetiness of the shared array is ensured
        (cf. multiprocessing.Array) through the use of
        SharedNumpyArray.acquire() and .release(),
        but this is of the responsability of the user implementing the
        Worker's inner task.
        """
        import numpy
        if isinstance(array, sharedctypes.SynchronizedArray):
            self._mp_array = array
            self._np_array = numpy.frombuffer(array.get_obj())
        elif (isinstance(array, numpy.ndarray) or
              isinstance(array, numpy.ma.masked_array)):
            if array.dtype in (float, numpy.float64):
                typecode = 'd'
            elif array.dtype in (int, numpy.int64):
                typecode = 'l'
            elif array.dtype in (numpy.int32,):
                typecode = 'i'
            elif array.dtype in (numpy.float32,):
                typecode = 'f'
            else:
                raise NotImplementedError('array.dtype==' + str(array.dtype))
            self._mp_array = multiprocessing.Array(typecode, array.size)
            np_array = numpy.frombuffer(self._mp_array.get_obj(), dtype=array.dtype)
            self._np_array = np_array.reshape(array.shape)
            self._np_array[...] = array[...]
            if isinstance(array, numpy.ma.masked_array):
                self._np_array = numpy.ma.masked_where(array.mask, self._np_array, copy=False)

    # redefinition of implicit methods: hash because of Footprints, eq to be
    # redefined together with hash in Python3 and transitivity of numpy arrays
    # methods from the inner numpy array layer
    def __hash__(self):
        return hash(self._mp_array)

    def __eq__(self, other):
        if isinstance(other, SharedNumpyArray):
            return self._np_array == other._np_array
        else:
            return False

    def __getattribute__(self, attr):
        if attr in ('_np_array', '_mp_array'):
            return super().__getattribute__(attr)
        elif attr in ('get_lock', 'acquire', 'release'):
            return self._mp_array.__getattribute__(attr)
        else:
            if hasattr(self._np_array, attr):
                return self._np_array.__getattribute__(attr)
            else:
                raise AttributeError(attr)

    def __setattr__(self, attr, value):
        if attr in ('_np_array', '_mp_array'):
            super().__setattr__(attr, value)
        else:
            self._np_array.__setattr__(attr, value)

    def __getitem__(self, *args):
        return self._np_array.__getitem__(*args)

    def __setitem__(self, *args):
        self._np_array.__setitem__(*args)

    def __setslice(self, *args):
        self._np_array.__setslice__(*args)

    def __getslice__(self, *args):
        return self._np_array.__getslice__(*args)



import sys
import time

from unittest import TestCase, main, skipIf

import footprints
import taylorism
from taylorism import examples, schedulers, taylorism_log
from bronx.fancies import loggers
from bronx.system import interrupt  # because subprocesses must be killable properly
from bronx.system import cpus as cpus_tool

try:
    import numpy as np
except ImportError:
    numpy_looks_fine = False
else:
    numpy_looks_fine = True

tloglevel = 'CRITICAL'
tloglevel_taylorism = 'CRITICAL'


def stderr2out_deco(f):
    def wrapped_f(*kargs, **kwargs):
        oldstderr = sys.stderr
        sys.stderr = sys.stdout
        try:
            return f(*kargs, **kwargs)
        finally:
            sys.stderr = oldstderr
    wrapped_f.__name__ = f.__name__
    return wrapped_f


class _TestError(Exception):
    pass


class Succeeder(examples.Sleeper):
    """Does nothing, but succeeds at it."""

    _footprint = dict(
        priority=dict(
            level=footprints.priorities.top.level('debug')
        ),
        info="Suceeds.",
        attr=dict(
            succeed=dict(
                info="Supposed to succeed.",
                type=bool,
                values=[True]
            ),
            bind_test=dict(
                info="Do the bind test.",
                type=bool,
                optional=True,
                default=False,
            )
        )
    )

    def _task(self):
        """Succeed at doing nothing."""
        time.sleep(self.sleeping_time)
        if self.bind_test:
            return ("Succeeded.", self.binding())
        else:
            return ("Succeeded.", )


class Failer(examples.Sleeper):
    """Does nothing, but fails at it."""

    _footprint = dict(
        priority=dict(
            level=footprints.priorities.top.level('debug')
        ),
        info="Fails.",
        attr=dict(
            succeed=dict(
                info="Supposed to fail.",
                type=bool,
                values=[False]
            ),
        )
    )

    def _task(self):
        """Fails (an exception is raised) at doing nothing."""
        time.sleep(self.sleeping_time)
        raise _TestError("Failer: failed")


@loggers.unittestGlobalLevel(tloglevel)
class UtTaylorism(TestCase):

    @stderr2out_deco
    def test_worker_met_an_exception(self):
        """
        Run a Succeeder and a Failer, checks that the Failer exception is
        catched.
        """
        taylorism_log.setLevel(tloglevel_taylorism)
        boss = taylorism.run_as_server(
            common_instructions=dict(),
            individual_instructions=dict(sleeping_time=[0.001, 0.01], succeed=[False, True]),
            scheduler=footprints.proxy.scheduler(limit='threads', max_threads=2),
        )
        with interrupt.SignalInterruptHandler(emitlogs=False):
            with self.assertRaises(_TestError):
                boss.wait_till_finished()

    @stderr2out_deco
    def test_boss_crashes(self):
        """
        Run a Succeeder and a Failer, checks that an error in the Boss
        subprocess is catched.
        """
        taylorism_log.setLevel(tloglevel_taylorism)
        boss = taylorism.run_as_server(
            common_instructions=dict(),
            individual_instructions=dict(sleeping_time=[60, 60], succeed=[True, True]),
            scheduler=footprints.proxy.scheduler(limit='threads', max_threads=2),
        )
        with interrupt.SignalInterruptHandler(emitlogs=False):
            with self.assertRaises(interrupt.SignalInterruptError):
                time.sleep(2)
                boss._process.terminate()
                boss.wait_till_finished()

    def test_servermode(self):
        """Run as server mode, checks appending instructions."""
        # Test both new and legacy schedulers
        taylorism_log.setLevel(tloglevel_taylorism)
        for scheduler in (footprints.proxy.scheduler(limit='threads', max_threads=2),
                          schedulers.MaxThreadsScheduler(max_threads=2)):
            boss = taylorism.run_as_server(
                common_instructions=dict(succeed=True,),
                individual_instructions=dict(sleeping_time=[0.001, 0.001, 0.001]),
                scheduler=scheduler,
            )
            time.sleep(0.2)
            boss.set_instructions(dict(succeed=True,),
                                  individual_instructions=dict(sleeping_time=[0.001, ]))
            boss.wait_till_finished()
            report = boss.get_report()
            self.assertEqual(len(report['workers_report']),
                             4,
                             "4 instructions have been sent, which is not the size of report.")

    @stderr2out_deco
    def test_toomany_instr_after_crash(self):
        """
        Checks that overloading the instructions queue after end of
        subprocess does not lead to deadlock.
        """
        taylorism_log.setLevel(tloglevel_taylorism)
        boss = taylorism.run_as_server(
            common_instructions=dict(),
            individual_instructions=dict(sleeping_time=[0.001, 60], succeed=[False, True]),
            scheduler=footprints.proxy.scheduler(limit='threads', max_threads=2),
        )
        time.sleep(0.5)
        with interrupt.SignalInterruptHandler(emitlogs=False):
            with self.assertRaises(_TestError):
                boss.set_instructions(
                    dict(),
                    individual_instructions=dict(sleeping_time=[1, ], bidon=['a' * 100000000, ])
                )

    def test_binding(self):
        """Checks that the binding works."""
        try:
            li = cpus_tool.LinuxCpusInfo()
            avcpus = cpus_tool.get_affinity()
        except cpus_tool.CpusToolUnavailableError as e:
            raise self.skipTest(str(e))
        if set(li.cpus.keys()) != avcpus:
            raise self.skipTest('The host is not entirely available.')
        taylorism_log.setLevel(tloglevel_taylorism)
        boss = taylorism.run_as_server(
            common_instructions=dict(wakeup_sentence='yo', succeed=True, bind_test=True),
            individual_instructions=dict(sleeping_time=[0.001, 0.001, 0.001]),
            scheduler=footprints.proxy.scheduler(limit='threads', max_threads=2, binded=True),
        )
        boss.wait_till_finished()
        report = boss.get_report()
        self.assertEqual(len(report['workers_report']), 3,
                         "3 instructions have been sent, which is not the size of report.")
        self.assertEqual({r['report'][1][0] for r in report['workers_report']},
                         set(list(li.socketpacked_cpulist())[:2]))

    @stderr2out_deco
    def test_redundant_workers_name(self):
        """
        Checks that a clear error is raised if several workers wear the same
        name.
        """
        taylorism_log.setLevel(tloglevel_taylorism)
        with self.assertRaises(ValueError):
            boss = taylorism.run_as_server(
                common_instructions=dict(),
                individual_instructions=dict(name=['alfred', 'alfred'],
                                             sleeping_time=[60, 60],
                                             succeed=[True, True]),
                scheduler=footprints.proxy.scheduler(limit='threads', max_threads=2),
            )
            boss.wait_till_finished()

    def test_expansion_workers_name(self):
        """Checks that expansion in workers name works fine."""
        taylorism_log.setLevel(tloglevel_taylorism)
        boss = taylorism.run_as_server(
            common_instructions=dict(name='jean-pierre_[sleeping_time]', succeed=True,),
            individual_instructions=dict(sleeping_time=[0.001, 0.01]),
            scheduler=footprints.proxy.scheduler(limit='threads', max_threads=2),
        )
        boss.wait_till_finished()
        report = boss.get_report()
        self.assertEqual(len(report['workers_report']),
                         2,
                         "2 instructions have been sent, which is not the size of report.")

    @skipIf(not numpy_looks_fine, "NumPy is unavailable.")
    def test_sharedmemory_array(self):
        """Checks that sharedmemory mechanism works fine."""
        taylorism_log.setLevel(tloglevel_taylorism)
        vals = [813, 42, 8]
        s = taylorism.util.SharedNumpyArray(np.ones((1,), dtype=int) * vals[0])
        boss = taylorism.run_as_server(
            common_instructions=dict(use_lock=True),
            individual_instructions=dict(value=vals[1:]),
            scheduler=footprints.proxy.scheduler(limit='threads', max_threads=2),
            sharedmemory_common_instructions=dict(shared_sum=s)
        )
        boss.wait_till_finished()
        self.assertEqual(s[0],
                         sum(vals),
                         "sharedmemory array has wrong value:{} instead of expected: {}."
                         .format(s[0], sum(vals)))


if __name__ == '__main__':
    main(verbosity=2)



"""
Contains classes for Schedulers.

Among a set of instructions to be passed to a Worker, and according to its own
criteria, the Scheduler determine at the current moment the ones that can be
launched right now simultaneously, and those that must be delayed.

A scheduler hence basically has one method:
launchable(pending_instructions, workers, report).

Its parameters (constant among execution) can be attributed in its constructor.
Other quantities, variables among execution, must be available within
*workers* (work being done) and *report* (work done).

A set of basic schedulers is given.

Starting from version 1.0.7, schedulers should be created using the footprints
package::

    import footprints as fp
    # In order to create a NewMaxThreadsScheduler scheduler:
    mt_sched = fp.proxy.scheduler(limit='threads', max_threads=2)

Compatibility classes are still provided (see :class:`LaxistScheduler`,
:class:`MaxThreadsScheduler`, :class:`MaxMemoryScheduler` and
:class:`SingleOpenFileScheduler`) but they should not be used anymore.

Dependencies
------------

:mod:`footprints` (MF package)
"""

import footprints
from footprints import FootprintBase
from bronx.fancies import loggers
from bronx.system import cpus, memory
from bronx.syntax.decorators import secure_getattr

import multiprocessing

logger = loggers.getLogger(__name__)

MAX_NUMBER_PROCESSES = 512


class BaseScheduler(FootprintBase):
    """Abstract base class for schedulers."""

    _abstract = True
    _collector = ('scheduler',)
    _footprint = dict(
        attr=dict(
            identity=dict(
                info="Scheduler identity.",
                optional=True,
                default='anonymous',
            ),
        )
    )

    def launchable(self, pending_instructions, workers, report):
        """
        Split *pending_instructions* into "launchable" and "not_yet_launchable"
        instructions according to the scheduler own rules.

        For that purpose and in a generic manner, the scheduler may need:

        - *pending_instructions*: todo
        - *workers*: being done
        - *report*: done.
        """
        raise NotImplementedError('launchable() method must be implemented in \
                                   inheritant classes. (BaseScheduler is abstract).')

    def _all_tickets(self):
        return {None}

    def _workers_hooks(self):
        """Return a list of callbacks to be triggered before workers task processing."""
        return list()

    def _assign_tickets(self, workers, launchable):
        """Assign available tickets in **launchable** instructions."""
        assigned_tickets = {w.scheduler_ticket for w in workers.values()}
        possible_tickets = sorted(self._all_tickets() - assigned_tickets)
        for instructions in launchable:
            possible_tickets.append(None)
            instructions.update(
                scheduler_ticket=possible_tickets.pop(0),
                scheduler_hooks=self._workers_hooks(),
            )
        return launchable


class NewLaxistScheduler(BaseScheduler):
    """No sorting is done !"""

    _footprint = dict(
        attr=dict(
            nosort=dict(
                alias=('laxist', 'unsorted'),
                values=(True,),
                type=bool,
            ),
        )
    )

    def launchable(self, pending_instructions, workers, report):
        """Very crude strategy: any pending instruction could be triggered."""
        launchable = self._assign_tickets(workers, pending_instructions)
        return launchable, list()


class NewLimitedScheduler(BaseScheduler):
    """
    A scheduler that dequeue the pending list as long as a maximum number
    of simultaneous tasks (*max_threads*) is not reached.
    """

    _abstract = True,
    _footprint = dict(
        attr=dict(
            limit=dict(
                values=['threads', 'memory', 'mem', 'threads+memory'],
                remap=dict(mem='memory'),
            ),
        )
    )


#: Abstract footprint attribute for binding aware schedulers
_binded_fpattr = footprints.Footprint(info='Abstract binded attribute',
                                      attr=dict(binded=dict(type=bool,
                                                            info="Binds the process to a single cpu.",
                                                            default=False,
                                                            optional=True)))


def binding_setup(worker):
    """Bind a *worker* to its *scheduler_ticket* compute core."""
    cpusinfo = cpus.LinuxCpusInfo()
    cpuslist = list(cpusinfo.socketpacked_cpulist())
    binded_cpu = cpuslist[worker.scheduler_ticket % cpusinfo.nvirtual_cores]
    cpus.set_affinity(binded_cpu)


def BindingAwareScheduler(cls):
    """
    A class decorator that wraps the original _workers_hooks method to add
    binding's setup method to the list Worker's hooks.

    NB: The class' footprint should include a 'binded' attribute.
    """
    # Wrap _workers_hooks
    original_hooks = getattr(cls, '_workers_hooks')

    def new_hooks(self):
        hookslist = original_hooks(self)
        if getattr(self, 'binded', False):
            hookslist.append(binding_setup)
        return hookslist

    cls._workers_hooks = new_hooks
    return cls


@BindingAwareScheduler
class NewMaxThreadsScheduler(NewLimitedScheduler):
    """
    A basic scheduler that dequeue the pending list as long as a maximum number
    of simultaneous tasks (*max_threads*) is not reached.
    """

    _footprint = [
        _binded_fpattr,
        dict(
            attr=dict(
                limit=dict(
                    values=['threads', 'processes'],
                    remap=dict(processes='threads'),
                ),
                max_threads=dict(
                    alias=('maxpc', 'maxthreads'),
                    remap={0: multiprocessing.cpu_count() / 2},
                    type=int,
                ),
            )
        )
    ]

    def _all_tickets(self):
        """The actual range of available tickets is limited by a maximum number of threads."""
        return set(range(0, self.max_threads))

    def launchable(self, pending_instructions, workers, report):
        """Limit strategy: only max_threads processes could run simultaneously."""
        available_threads = self.max_threads - len(workers)
        launchable = pending_instructions[0:max(available_threads, 0)]
        not_yet_launchable = pending_instructions[max(available_threads, 0):]
        launchable = self._assign_tickets(workers, launchable)
        return launchable, not_yet_launchable


@BindingAwareScheduler
class NewMaxMemoryScheduler(NewLimitedScheduler):
    """
    A basic scheduler that dequeue the pending list as long as a critical memory
    level (according to 'memory' element of workers instructions (in MB) and
    total system memory) is not reached.
    """

    _footprint = [
        _binded_fpattr,
        dict(
            attr=dict(
                limit=dict(
                    values=['memory', 'mem'],
                    remap=dict(mem='memory'),
                ),
                max_memory=dict(
                    info="Amount of usable memroy (in MiB)",
                    optional=True,
                    type=float,
                    access='rwx',
                ),
                memory_per_task=dict(
                    info=("If a worker do not provide any information on memory, " +
                          "request at least *memory_per_task* MiB of memory."),
                    optional=True,
                    default=2048.,
                    type=float,
                ),
                memory_max_percentage=dict(
                    info=("Max memory level as a fraction of the total" +
                          "system memory (used only if max_memroy is not provided)."),
                    optional=True,
                    default=0.75,
                    type=float,
                ),
            )
        )
    ]

    def __init__(self, *args, **kw):
        """Setup the maximum available memory."""
        super().__init__(*args, **kw)
        if self.max_memory is None:
            # memory tools are all but generic, they might fail !
            try:
                system_mem = memory.LinuxMemInfo().system_RAM('MiB')
            except OSError:
                raise OSError("Unable to determine the total system's memory size.")
            self.max_memory = self.memory_max_percentage * system_mem

    def _all_tickets(self):
        """The actual range of available tickets is limited by a maximum number of threads."""
        return set(range(0, MAX_NUMBER_PROCESSES))

    def launchable(self, pending_instructions, workers, report):
        """Limit strategy: only processes that fit in a given amount of memory could run."""
        used_memory = sum([w.memory or self.memory_per_task for w in workers.values()])
        launchable = list()
        not_yet_launchable = list()
        for instructions in pending_instructions:
            actual_memory = instructions.get('memory', self.memory_per_task)
            if used_memory + actual_memory < self.max_memory:
                launchable.append(instructions)
                used_memory += actual_memory
            else:
                not_yet_launchable.append(instructions)
        launchable = self._assign_tickets(workers, launchable)
        return launchable, not_yet_launchable


class LongerFirstScheduler(NewMaxMemoryScheduler):
    """
    A scheduler based on the NewMaxMemory Scheduler. It aims at launching as soon
    as possible the workers that are expected to have the longest run-time (whilst
    there is enough available memory).

    Workers needs to have 2 attributes:

        * expected_time representing the expected run time
        * memory representing the expected memory consumed

    This scheduler is typically used with Bateur workers, in vortex's
    src/common/algo/odbtools.py (for parallel BATOR run).
    """

    _footprint = dict(
        attr=dict(
            limit=dict(
                values=['threads+memory'],
            ),
            max_threads=dict(
                alias=('maxpc', 'maxthreads'),
                remap={0: multiprocessing.cpu_count() / 2},
                type=int,
            ),
        )
    )

    def _all_tickets(self):
        """The actual range of available tickets is limited by a maximum number of threads."""
        return set(range(0, self.max_threads))

    def launchable(self, pending_instructions, workers, report):
        """Limit strategy: only processes that fit in a given amount of memory could run."""
        available_threads = self.max_threads - len(workers)
        used_memory = sum([w.memory or self.memory_per_task for w in workers.values()])
        available_memory = self.max_memory - used_memory
        launchable = []
        not_yet_launchable = []
        # sort by decreasing time
        pending_instructions.sort(key=lambda tup: tup.get('expected_time', 0), reverse=True)
        for instructions in pending_instructions:
            actual_memory = instructions.get('memory', self.memory_per_task)
            if (actual_memory <= available_memory) and (available_threads > 0):
                launchable.append(instructions)
                available_memory -= actual_memory
                available_threads -= 1
            else:
                not_yet_launchable.append(instructions)

        launchable = self._assign_tickets(workers, launchable)
        return launchable, not_yet_launchable


class NewSingleOpenFileScheduler(NewMaxThreadsScheduler):
    """
    Ensure that files will not be open 2 times simultaneously by 2 workers.
    And with a maximum threads number.
    """

    _footprint = dict(
        attr=dict(
            singlefile=dict(
                values=(True,),
                type=bool,
            ),
        )
    )

    def launchable(self, pending_instructions, workers, report):
        """Limit strategy: deal with what looks like a file locking."""
        launchable = []
        not_yet_launchable = []
        open_files = [w.fileA for w in workers.values()] + [w.fileB for w in workers.values()]
        # check files are not already open by other worker
        for pi in pending_instructions:
            if pi['fileA'] in open_files or pi['fileB'] in open_files:
                not_yet_launchable.append(pi)
            else:
                launchable.append(pi)
        # and finally sort with regards to MaxThreads
        (launchable, nyl) = super().launchable(launchable, workers, report)
        not_yet_launchable.extend(nyl)
        launchable = self._assign_tickets(workers, launchable)
        return launchable, not_yet_launchable


# ------------------------------------------------------------------------------
# The following classes are kept for backward compatibility. From now and on, one
# should abstain to use them.


class _AbstractOldSchedulerProxy:
    """the abstract class of deprecated scheduler objects."""

    _TARGET_CLASS = None

    def __init__(self, *kargs, **kwargs):
        if self._TARGET_CLASS is None:
            raise RuntimeError('_TARGET_CLASS needs to be defined')
        logger.warning('The %s class is deprecated. ' +
                       'Instead, use the footprint mechanism to create schedulers.',
                       self.__class__.__name__)
        self.__target_scheduler = self._TARGET_CLASS(*kargs, **kwargs)
        super().__init__()

    @secure_getattr
    def __getattr__(self, name):
        return getattr(self.__target_scheduler, name)


class LaxistScheduler(_AbstractOldSchedulerProxy):
    """Deprecated class: should not be used from now and on."""

    _TARGET_CLASS = NewLaxistScheduler

    def __init__(self):
        super().__init__(nosort=True)


class MaxThreadsScheduler(_AbstractOldSchedulerProxy):
    """Deprecated class: should not be used from now and on."""

    _TARGET_CLASS = NewMaxThreadsScheduler

    def __init__(self, max_threads=0):
        super().__init__(limit='threads', max_threads=max_threads)


class MaxMemoryScheduler(_AbstractOldSchedulerProxy):
    """Deprecated class: should not be used from now and on."""

    _TARGET_CLASS = NewMaxMemoryScheduler

    def __init__(self, max_memory_percentage=0.75, total_system_memory=None):
        """
        :param float max_memory_percentage: Max memory level as a fraction of the total system memory
        :param float total_system_memory: Memory available on this system (in MiB)
        """
        if total_system_memory is not None:
            max_memory = total_system_memory * max_memory_percentage
        else:
            max_memory = None
        super().__init__(limit='memory',
                         max_memory=max_memory,
                         memory_max_percentage=max_memory_percentage)


class SingleOpenFileScheduler(_AbstractOldSchedulerProxy):
    """Deprecated class: should not be used from now and on."""

    _TARGET_CLASS = NewSingleOpenFileScheduler

    def __init__(self, max_threads=0):
        super().__init__(limit='threads', max_threads=max_threads, singlefile=True)



"""
Framework for parallelisation of tasks.
"""

import multiprocessing as mpc
import queue
import uuid
import sys
import traceback
import copy
import os
from pickle import PickleError

from footprints import FootprintBase, FPList, proxy as fpx
from bronx.fancies import loggers
from bronx.system import interrupt, cpus  # because subprocesses must be killable properly

from .schedulers import BaseScheduler
from .schedulers import MaxThreadsScheduler, binding_setup  # For compatibility

assert BaseScheduler
assert MaxThreadsScheduler

taylorism_log = loggers.getLogger(__name__)

# : timeout when polling for a Queue/Pipe communication
communications_timeout = 0.01


# FUNCTIONS
###########

def run_as_server(common_instructions=dict(),
                  individual_instructions=dict(),
                  scheduler=None,
                  verbose=False,
                  maxlenreport=1024,
                  sharedmemory_common_instructions=dict()):
    """
    Build a Boss instance, make him hire workers,
    run the workers, and returns the Boss instance.

    Be aware that the Boss MUST be told when no more instructions will be
    appended, or the subprocess will continue to live alone (until
    destruction of the Boss instance).

    :param dict common_instructions: to be passed to the workers
    :param dict individual_instructions: to be passed to the workers
    :param scheduler: scheduler to rule scheduling of workers/threads
    :param bool verbose: is the Boss verbose or not.
    :param int maxlenreport: the maximum number of lines for the report (when
        running in verbose mode)
    :param dict sharedmemory_common_instructions: special "instructions", whose
        memory allocation is shared among workers and from main process.
        Warning: these objects must inherit in some way from
        multiprocessing.Array or sharedctypes. For n-dimensional arrays,
        it is advised to be instances of the here-defined :class:`SharedNumpyArray`.
    """
    if scheduler is None:
        scheduler = fpx.scheduler(limit='threads', max_threads=0)
    boss = Boss(verbose=verbose, scheduler=scheduler, maxlenreport=maxlenreport,
                sharedmemory_common_instructions=sharedmemory_common_instructions)
    boss.set_instructions(common_instructions, individual_instructions)
    boss.make_them_work()
    return boss


def batch_main(common_instructions=dict(),
               individual_instructions=dict(),
               scheduler=None,
               verbose=False,
               maxlenreport=1024,
               print_report=print,
               sharedmemory_common_instructions=dict()):
    """
    Run execution of the instructions as a batch process, waiting for all
    instructions are finished and finally printing report.

    Args and kwargs are those of run_as_server() function.
    """
    if scheduler is None:
        scheduler = fpx.scheduler(limit='threads', max_threads=0)
    boss = run_as_server(common_instructions,
                         individual_instructions,
                         scheduler=scheduler,
                         verbose=verbose,
                         maxlenreport=maxlenreport,
                         sharedmemory_common_instructions=sharedmemory_common_instructions)

    with interrupt.SignalInterruptHandler(emitlogs=False):
        try:
            boss.wait_till_finished()
            report = boss.get_report()
        except (Exception, KeyboardInterrupt, interrupt.SignalInterruptError):
            boss.stop_them_working()
            boss.wait_till_finished()
            raise
        else:
            for r in report['workers_report']:
                taskheader = 'WORKER NAME: ' + r['name']
                print('=' * len(taskheader))
                print(taskheader)
                print('-' * len(taskheader))
                print_report(r['report'])
            return report


# MAIN CLASSES
##############
class Worker(FootprintBase):
    """
    Template for workers.
    A Worker is an object supposed to do a task, according to instructions.
    The instructions has to be added to footprint attributes in actual classes.
    """

    _abstract = True
    _collector = ('worker',)
    _footprint = dict(
        attr=dict(
            name=dict(
                info='Name of the worker.',
                optional=True,
                default=None,
                access='rwx',
            ),
            memory=dict(
                info='Memory that should be used by the worker (in MiB).',
                optional=True,
                default=0.,
                type=float,
            ),
            expected_time=dict(
                info='How long the worker is expected to run (in s).',
                optional=True,
                default=0.,
                type=float,
            ),
            scheduler_ticket=dict(
                info='The slot number given by the scheduler (optional).',
                optional=True,
                default=None,
                type=int,
            ),
            scheduler_hooks=dict(
                info='List of callbacks before starting effective task work.',
                optional=True,
                default=FPList(),
                type=FPList,
            ),
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.name is None:
            self.name = str(uuid.uuid4())
            taylorism_log.debug("Worker's name auto-assigned: %s", self.name)
        self._parent_pid = os.getpid()
        self._process = mpc.Process(target=self._work_and_communicate)
        self._terminating = False
        self._messenger = None

    def __del__(self):
        if (hasattr(self, '_process') and self._process.pid and
                # A subprocess should never call join on itself...
                self._parent_pid == os.getpid()):
            self._process.join(0.1)
            if self._process.is_alive():
                self._process.terminate()
                taylorism_log.debug('Worker process terminate issued (in __del__): pid=%s. name=%s',
                                    self._process.pid, self.name)

    def _get_messenger(self):
        """Return actual messenger for communication with the boss."""
        return self._messenger

    def _set_messenger(self, messenger):
        """Connect to some Queue."""
        assert callable(messenger.put)
        self._messenger = messenger

    messenger = property(_get_messenger, _set_messenger)

    def binding(self):
        """Return the actual physical binding of the current process to a cpu if available.

        The :class:`cpus.CpusToolUnavailableError` may be raised depending
        on the system.
        """
        cpuloc = cpus.get_affinity()
        if cpuloc == set(cpus.LinuxCpusInfo().raw_cpulist()):
            cpuloc = [None]
        return list(cpuloc)

    def work(self):
        """Send the Worker to his job."""
        if not self._terminating:
            self._process.start()
            taylorism_log.debug('Worker process started: pid=%s. name=%s',
                                self._process.pid, self.name)
        else:
            taylorism_log.debug('Worker process cannot be started while terminating. name=%s.',
                                self.name)

    def bye(self):
        """
        Block the Boss until the worker has finished his job.
        THIS METHOD SHOULD NEVER BE CALLED BY THE OBJECT HIMSELF !
        (WOULD CAUSE A DEADLOCK if called from inside the worker's subprocess)
        """
        if self._process.pid:
            self._process.join()
            taylorism_log.debug('Worker process joined: pid=%s. name=%s',
                                self._process.pid, self.name)
        else:
            taylorism_log.debug('Worker process not yet started. Nothing to do. name=%s.',
                                self.name)

    def stop_working(self):
        """Make the worker stop working.

        Since the worker process sets up a Signal handler, it should not stop
        abruptly when this method is called for the first time...
        """
        if not self._terminating:
            if self._process.pid:
                self._process.terminate()
                taylorism_log.debug('Worker process terminated (#1): pid=%s. name=%s',
                                    self._process.pid, self.name)
            else:
                taylorism_log.debug('Worker process not yet started. Nothing to do. name=%s.',
                                    self.name)
            self._terminating = True
        else:
            if self._process.pid:
                self._process.join(0.1)
                taylorism_log.debug('Worker process joined: pid=%s. name=%s',
                                    self._process.pid, self.name)
                if self._process.is_alive():
                    self._process.terminate()
                    taylorism_log.debug('Worker process terminated (#2): pid=%s. name=%s',
                                        self._process.pid, self.name)

    def _work_and_communicate(self):
        """
        Send the Worker to his task, making sure he communicates with
        its boss.

        From within this method down, everything is done in the subprocess
        world !
        """
        with interrupt.SignalInterruptHandler(emitlogs=False):
            fast_exit = False
            to_be_sent_back = dict(name=self.name, report=None)
            try:
                for callback in self.scheduler_hooks:
                    callback(self)
                self._work_and_communicate_prehook()
                to_be_sent_back.update(report=self._task())
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                tb = traceback.format_exception(exc_type,
                                                exc_value,
                                                exc_traceback)
                to_be_sent_back.update(report=e, traceback=tb)
            except (KeyboardInterrupt, interrupt.SignalInterruptError):
                fast_exit = True
            finally:
                if not fast_exit:
                    try:
                        self.messenger.put(to_be_sent_back)
                    except (ValueError, PickleError) as e:
                        # ValueError = to_be_sent_back too big.
                        # PickleError = to_be_sent_back unpickelable.
                        sys.stderr.write("The to_be_sent_back variable was:\n")
                        sys.stderr.write(str(to_be_sent_back))
                        to_be_sent_back.update(report=e, traceback='Traceback missing')
                        self.messenger.put(to_be_sent_back)
                self.messenger.close()

    def _work_and_communicate_prehook(self):
        """
        Some stuff executed before the "real" work_end_communicate takes place.
        """
        pass

    def _task(self, **kwargs):
        """
        Actual task of the Worker to be implemented therein.
        Return the report to be sent back to the Boss.
        """
        raise RuntimeError("this method must be implemented in Worker's inheritant class !")


class BindedWorker(Worker):
    """Workers binded to a cpu core (Linux only).

    This class is deprecated. Instead, inherit from :class:`Worker` and create
    a scheduler with binded=True.
    """

    _abstract = True

    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)
        taylorism_log.warning('The %s class is deprecated. Please use "Worker" instead.',
                              self.__class__.__name__)

    def _work_and_communicate_prehook(self):
        """Bind the process to a cpu"""
        if self.scheduler_ticket is not None:
            binding_setup(self)


class Boss:
    """
    Template for bosses.
    A Boss is an object supposed to order tasks to a series of workers.

    Optionally can be attributed to the Boss a *name* and a *verbose*-ity
    (to report in log, the workers reports).

    Also, a *scheduler* can be assigned, to rule the ordering of tasks to
    workers.
    Custom schedulers can be used, they only need to inherit from
    .schedulers.BaseScheduler and implemented launchable() method.
    """

    control_signals = {'HALT': 'Suspend ordering workers to work until RESUME.',
                       'RESUME': 'Resume loop on pending_instructions/workers.',
                       'SEND_REPORT': 'Send interim report (and continue normally).',
                       'END': 'Terminate all pending work, then Stop listening.\
                               No new instructions from control will be listened,\
                               except a STOP*.',
                       'STOP': 'Halt pending work, but let workers finish their\
                                current work, and then stop listening.',
                       'STOP_LISTENING': 'Stop listening, while workers continue\
                                          their current job.',
                       'STOP_RIGHTNOW': 'Stop workers immediately and stop\
                                         listening.'}

    def __init__(self, scheduler=None, name=None, verbose=False,
                 maxlenreport=1024, sharedmemory_common_instructions=dict()):
        if scheduler is None:
            scheduler = fpx.scheduler(limit='threads', max_threads=0)
        # Duck typing check...
        assert hasattr(scheduler, 'launchable')
        assert callable(scheduler.launchable)
        self.scheduler = scheduler
        self.name = name
        self.verbose = verbose
        self.maxlenreport = int(maxlenreport)
        self._sharedmemory_common_instructions = sharedmemory_common_instructions

        self.workers_messenger = mpc.Queue()
        (self.control_messenger_in,
         self.control_messenger_out) = mpc.Pipe()  # in = inside subprocess, out = main
        self.control_messenger_out.send(self.control_signals['HALT'])

        self._parent_pid = os.getpid()
        self._process = mpc.Process(target=self._listen_and_communicate)
        self._process.start()
        taylorism_log.debug('Boss process started: pid=%s. name=%s',
                            self._process.pid, str(self.name))

        self._finalreport = None

    def __del__(self):
        if (hasattr(self, '_process') and self._process.pid and
                # A subprocess should never call join on itself...
                self._parent_pid == os.getpid() and
                self._process.is_alive()):
            self._process.join(0.1)
            taylorism_log.debug('Boss process joined (in __del__): pid=%s. name=%s',
                                self._process.pid, str(self.name))
            if self._process.is_alive():
                self._process.terminate()
                taylorism_log.debug('Boss process terminated (in __del__): pid=%s. name=%s',
                                    self._process.pid, str(self.name))
        self.control_messenger_in.close()
        self.control_messenger_out.close()
        self.workers_messenger.close()

    def set_instructions(self,
                         common_instructions=dict(),
                         individual_instructions=dict(),
                         fatal=True):
        """
        Set instructions to be distributed to workers.

        :param dict common_instructions: are a series of arguments shared by
                                         each worker, to be passed to the Worker
                                         factory.
        :param dict individual_instructions: are a series of arguments proper to
                                             each worker, hence all individual
                                             instructions must have the same
                                             length
        :param bool fatal: if True, an error in parsing instructions will stop
                           the workers already running and the boss internal
                           subprocess
        """
        # parse instructions
        individual_instructions = copy.deepcopy(individual_instructions)
        instructions_sets = []
        if len(individual_instructions) > 0:
            # check their length is homogeneous
            _i0 = sorted(individual_instructions.keys())[0]
            indiv_instr_num = len(individual_instructions[_i0])  # length of first instruction
            if not all([len(instr) == indiv_instr_num
                        for instr in individual_instructions.values()]):
                if fatal:
                    self.stop_them_working()
                raise AssertionError("all *individual_instructions* must have the same length.")
            # gather common and individual
            for _ in range(indiv_instr_num):
                instructions = copy.copy(common_instructions)
                for k, v in individual_instructions.items():
                    instructions.update({k: v.pop(0)})
                instructions_sets.append(instructions)

        # send instructions to control
        if self._process.is_alive():
            try:
                self.control_messenger_out.send(instructions_sets)
            except (ValueError, PickleError):
                # ValueError = instructions_sets too big.
                # PickleError = instructions_sets unpickelable.
                sys.stderr.write("The instructions_sets variable was:\n")
                sys.stderr.write(str(instructions_sets))
                taylorism_log.error("Impossible to send data through the pipe")
                self.stop_them_working()
                raise
        else:
            self.get_report(interim=True)

    def make_them_work(self, terminate=False, stop_listening=False):
        """
        Order the workers to work.

        :param bool terminate: if True, no other instructions could be appended
                               later.
        :param bool stop_listening: if True, alive workers go on their jobs,
                                    but they are not listened to anymore;
                                    this is a bit tricky but might be useful ?
        """
        self.control_messenger_out.send(self.control_signals['RESUME'])
        if stop_listening:
            self.control_messenger_out.send(self.control_signals['STOP_LISTENING'])
        if terminate:
            self.end()

    def stop_them_working(self):
        """Stop the workers."""
        self.control_messenger_out.send(self.control_signals['STOP_RIGHTNOW'])

    def get_report(self, interim=True):
        """
        Get report of the work executed.

        If *interim*, ask for an interim report if no report is available,
        i.e. containing the work done by the time of calling.
        """
        return self._internal_get_report(interim=interim)

    def _internal_get_report(self, interim=True, final=False):
        """
        Get report of the work executed.

        :param bool interim: if True, ask for an interim report if no report is
                             available, i.e. containing the work done by the
                             time of calling.
        :param bool final: if True, the report is saved in an internal variable
                           and the saved report will be returned whenever
                           get_report is called.
        """
        received_a_report = self.control_messenger_out.poll

        def _getreport():
            if final or received_a_report():
                received = self._recv_report(splitmode=True)
                if final:
                    self._finalreport = received
                    self._process.join()
                    taylorism_log.debug('Boss process joined: pid=%s. name=%s',
                                        self._process.pid, str(self.name))
                if isinstance(received['workers_report'],
                              (Exception, KeyboardInterrupt, interrupt.SignalInterruptError)):
                    taylorism_log.error("Error was caught in subprocesses with traceback:")
                    sys.stderr.writelines(received['traceback'])
                    raise received['workers_report']
            else:
                received = None
            return received

        # first try to get report
        if self._finalreport is None:
            report = _getreport()
        else:
            report = self._finalreport

        if report is None and not final and interim:
            self.control_messenger_out.send(self.control_signals['SEND_REPORT'])
            while report is None:
                report = _getreport()
                if not self._process.is_alive():
                    break

        return report

    def end(self):
        """
        Ends the listening process once instructions are treated.
        MUST BE CALLED (or wait_till_finished) for each Boss to avoid zombies
        processes.
        """
        self.control_messenger_out.send(self.control_signals['END'])

    def wait_till_finished(self):
        """Block the calling tree until all instructions have been executed."""
        self.end()
        taylorism_log.debug('Boss process waiting for pending work: pid=%s. name=%s',
                            self._process.pid, str(self.name))
        self._internal_get_report(final=True)

# boss subprocess internal methods
##################################

    def _send_report(self, report, splitmode=True):
        """Report must have keys 'workers_report', 'status' and optionally others."""
        if not splitmode:
            self.control_messenger_in.send(report)
        else:
            rkeys = list(report.keys())
            rkeys.pop(rkeys.index('workers_report'))
            rkeys.pop(rkeys.index('status'))
            for k in rkeys:
                self.control_messenger_in.send((k, report[k]))
            if not isinstance(report['workers_report'],
                              (Exception, KeyboardInterrupt, interrupt.SignalInterruptError)):
                for wr in report['workers_report']:
                    self.control_messenger_in.send(wr)
            else:
                self.control_messenger_in.send(report['workers_report'])
            self.control_messenger_in.send(('status', report['status']))

    def _recv_report(self, splitmode=True):
        """Report must have keys 'workers_report', 'status' and optionally others."""
        if not splitmode:
            report = self.control_messenger_out.recv()
        else:
            report = {'workers_report': []}
            while True:
                r = self.control_messenger_out.recv()
                if isinstance(r, tuple):
                    report[r[0]] = r[1]
                    if r[0] == 'status':
                        break
                elif isinstance(r, dict):
                    report['workers_report'].append(r)
                elif isinstance(r, (Exception, KeyboardInterrupt, interrupt.SignalInterruptError)):
                    report['workers_report'] = r
        return report

    def _listen_and_communicate(self):
        """Interface routine, to catch exceptions and communicate.

        From within this method down, everything is done in the subprocess
        world !
        """
        with interrupt.SignalInterruptHandler(emitlogs=False):
            try:
                (workers_report, pending_instructions) = self._listen()
                if len(pending_instructions) == 0:
                    report = {'workers_report': workers_report,
                              'status': 'finished'}
                else:
                    report = {'workers_report': workers_report,
                              'status': 'pending',
                              'pending': pending_instructions}
            except (Exception, KeyboardInterrupt, interrupt.SignalInterruptError) as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                tb = traceback.format_exception(exc_type,
                                                exc_value,
                                                exc_traceback)
                report = {'workers_report': e,
                          'status': 'workers exception',
                          'traceback': tb}
            finally:
                try:
                    self._send_report(report, splitmode=True)
                except (ValueError, OSError) as e:
                    # ValueError = to_be_sent_back too big.
                    # We are sure that a PickleError won't occur since data were
                    # already pickled once (by the workers)
                    taylorism_log.error("The report is too big to be sent back :-(")
                    report = {'workers_report': e,
                              'status': 'transmission exception',
                              'traceback': 'Traceback missing'}
                    self._send_report(report, splitmode=True)

    def _stop_them_working(self, workers):
        # Issue the terminate signal (SIGTERM)
        for wname in list(workers.keys()):
            workers[wname].stop_working()
        # Empty the message queue (but do not process messages) because some
        # of the workers may have completed there work in the meantime...
        empty = False
        while not empty:
            try:
                self.workers_messenger.get(timeout=communications_timeout)
            except queue.Empty:
                empty = True
        # Try to join everybody
        for wname in list(workers.keys()):
            workers.pop(wname).stop_working()

    def _hire_worker(self, workers, instructions):
        w = fpx.worker(**instructions)
        if w is None:
            raise AttributeError("no adequate Worker was found with these instructions: " +
                                 str(instructions))
        w.messenger = self.workers_messenger
        if w.name not in workers.keys():
            workers[w.name] = w
        else:
            raise ValueError('several workers wear the same name: ' + w.name)
        return w

    def _listen(self):
        """
        Actual listening method, i.e. running subprocess at interface between
        main and workers.

        Infinite loop:
          - A. listen to control, for appending new instructions or control signals
          - B. listen to workers, to collect their reports and/or errors
          - C. assign work to workers
          - D. exit loop if any reason for
        """
        workers = {}
        pending_instructions = []
        report = []

        halt = False
        end = False
        stop = False
        try:
            while True:
                # A. listen to control
                if self.control_messenger_in.poll(communications_timeout):
                    control = self.control_messenger_in.recv()
                    if control in self.control_signals.values():
                        # received a control signal
                        if control == self.control_signals['SEND_REPORT']:
                            try:
                                self._send_report(
                                    {'workers_report': report, 'status': 'interim'},
                                    splitmode=True
                                )
                            except ValueError:
                                # ValueError = report too big.
                                # We are sure that a PickleError won't occur
                                # since data were already pickled once (by the
                                # workers)
                                taylorism_log.error("The report is too big to be sent back :-(")
                                raise
                        elif control == self.control_signals['HALT']:
                            halt = True
                        elif control == self.control_signals['RESUME']:
                            halt = False
                        elif (control in [self.control_signals[k]
                                          for k in self.control_signals.keys()
                                          if 'STOP' in k] or
                              control == self.control_signals['END']):
                            end = True
                            if control == self.control_signals['STOP_LISTENING']:
                                break  # leave out the infinite loop
                            elif control in (self.control_signals['STOP'],
                                             self.control_signals['STOP_RIGHTNOW']):
                                stop = True
                                if control == self.control_signals['STOP_RIGHTNOW']:
                                    self._stop_them_working(workers)

                    else:
                        # received new instructions
                        if not end:
                            # if an END or STOP signal has been received,
                            # new instructions are not listened to
                            if isinstance(control, list):
                                pending_instructions.extend(control)
                            elif isinstance(control, dict):
                                pending_instructions.append(control)
                # B. listen to workers
                try:
                    reported = self.workers_messenger.get(timeout=communications_timeout)
                except queue.Empty:
                    pass
                else:
                    # got a new message from workers !
                    report.append(reported)
                    if isinstance(
                        reported['report'],
                        (Exception, KeyboardInterrupt, interrupt.SignalInterruptError)
                    ):
                        # worker got an exception
                        taylorism_log.error("error encountered with worker " +
                                            reported['name'] +
                                            " with traceback:")
                        sys.stderr.writelines(reported['traceback'])
                        sys.stderr.write("Instructions of guilty worker:\n")
                        w = [repr(a) + '\n'
                             for a in sorted(workers[reported['name']].footprint_as_dict().items())
                             if a]
                        sys.stderr.writelines(w)
                        if isinstance(reported['report'], Exception):
                            # The  KeyboardInterrupt/interrupt.SignalInterruptError case
                            # is handled latter on in the overall try/except
                            self._stop_them_working(workers)
                        raise reported['report']
                    else:
                        # worker has finished
                        if self.verbose:
                            msglog = str(reported['report'])
                            if len(msglog) > self.maxlenreport:
                                msglog = msglog[:self.maxlenreport] + ' ...'
                            taylorism_log.info(msglog)
                    workers.pop(reported['name']).bye()
                # C. there is work to do and no STOP signal has been received: re-launch
                if len(pending_instructions) > 0 and not stop and not halt:
                    (launchable, not_yet_launchable) = self.scheduler.launchable(
                        pending_instructions,
                        workers=workers,
                        report=report
                    )
                    for instructions in launchable:
                        instructions_and_shared_memory = instructions.copy()
                        instructions_and_shared_memory.update(self._sharedmemory_common_instructions)
                        try:
                            w = self._hire_worker(workers, instructions_and_shared_memory)
                        except (AttributeError, ValueError):
                            self._stop_them_working(workers)
                            raise
                        w.work()
                        if self.verbose:
                            taylorism_log.info(' '.join(['Worker',
                                                         w.name,
                                                         'started.']))
                    pending_instructions = not_yet_launchable
                # D. should we stop now ?
                if end and (len(workers) == len(pending_instructions) == 0):
                    # a STOP signal has been received, all workers are done and
                    # no more pending instructions remain:
                    # we can leave out infinite loop
                    stop = True
                if stop:
                    break
        except (interrupt.SignalInterruptError, KeyboardInterrupt):
            self._stop_them_working(workers)
            raise

        return report, pending_instructions



"""
Basic examples of how to use the module.

A more advanced example of use can be found in ``epygram``'s epy_conv.py tool.
"""

import time

from footprints import proxy as fpx

from taylorism import Worker, run_as_server
from .util import SharedNumpyArray


class Sleeper(Worker):
    """
    Sample worker for tutorial or debugging purpose, that sleeps a given time.

    The over-loading of __init__() is not mandatory, but a possibility.
    """

    _footprint = dict(
        info="Sleeps.",
        attr=dict(
            sleeping_time=dict(
                info="Sleeping time in s.",
                values=[0.001, 0.01, 0.1] + list(range(10)) + list(range(10, 65, 5)),
                type=float,
            ),
            wakeup_sentence=dict(
                info="What to say after sleep.",
                optional=True,
                access='rwx',
                default='Hello !',
            ),
        )
    )

    def _task(self):
        """
        Actual task of the Sleeper is implemented therein.
        Return the report to be sent back to the Boss.
        """
        time.sleep(self.sleeping_time)
        return ' '.join([self.wakeup_sentence, 'Woke up after',
                         str(self.sleeping_time), "s sleep on cpu", str(self.binding())])


class Logarithmer(Worker):
    """
    Computes the logarithm of an array.

    OK, this would be much more efficient with numpy.log(array), but it is just
    an example of using shared numpy arrays among workers.
    """
    _footprint = dict(
        attr=dict(
            row=dict(
                info="Index of the row of the array on which the Worker is supposed to work.",
                type=int),
            array=dict(
                info="The shared-memory array on which to work on.",
                type=SharedNumpyArray,
                access='rwx')
        )
    )

    def _task(self):
        import numpy
        for j in range(self.array.shape[1]):
            self.array[self.row, j] = numpy.log(self.array[self.row, j])


class Summer(Worker):
    """
    Computes a Sum, each Worker adding its value to a SharedNumpyArray.

    This needs and illustrates process-safetiness.
    """
    _footprint = dict(
        attr=dict(
            value=dict(
                info="Value to be added by the worker.",
                type=int),
            shared_sum=dict(
                info="The shared-memory array on which to sum.",
                type=SharedNumpyArray,
                access='rwx'),
            use_lock=dict(
                info="Whether to use the lock (thread-safe) or not (may lead to a wrong result !).",
                type=bool)
        )
    )

    def _task(self):
        if self.use_lock:
            # acquire the lock, to be sure no other process is accessing the data meanwhile
            self.shared_sum.acquire()
        self.shared_sum[0] += self.value
        if self.use_lock:
            self.shared_sum.release()  # release the lock


class MatrixProducter(Worker):
    """
    Computes a Matrix Product C = A x B by hand, but parallel by blocks,
    using 3 SharedNumpyArray, each Worker being responsible for a [i1:i2, j1:j2]
    block of the resulting C matrix.

    Again, numpy matrix products may probably be more efficient...
    """
    _footprint = dict(
        attr=dict(
            A=dict(
                info="The A shared-memory array matrix.",
                type=SharedNumpyArray,
                access='rwx'),
            B=dict(
                info="The B shared-memory array matrix.",
                type=SharedNumpyArray,
                access='rwx'),
            C=dict(
                info="The C shared-memory array matrix.",
                type=SharedNumpyArray,
                access='rwx'),
            i1=dict(
                info="The first index i of the the Worker is responsible for.",
                type=int),
            i2=dict(
                info="The last index i of the the Worker is responsible for.",
                type=int),
            j1=dict(
                info="The first index j of the the Worker is responsible for.",
                type=int),
            j2=dict(
                info="The first index j of the the Worker is responsible for.",
                type=int),
        )
    )

    def _task(self):
        for i in range(self.i1, self.i2 + 1):
            for j in range(self.j1, self.j2 + 1):
                self.C[i, j] = sum([self.A[i, k] * self.B[k, j] for k in range(self.A.shape[1])])


def sleepers_generic_program(verbose=True, scheduler=None):
    """Generic example: how to run and control the Boss."""
    boss = run_as_server(
        common_instructions=dict(wakeup_sentence='Hello Dolly !'),
        individual_instructions=dict(sleeping_time=[4, 9, 2, 1]),
        scheduler=scheduler,
        verbose=verbose,
    )
    time.sleep(5)
    print('Intermediate report:', boss.get_report())
    boss.set_instructions(dict(), individual_instructions=dict(sleeping_time=[3, ]))
    boss.wait_till_finished()
    report = boss.get_report()
    for r in report['workers_report']:
        print(r)


def sleepers_example_laxist(verbose=True):
    """Example: assuming no selection of strategy for scheduling."""
    sleepers_generic_program(
        verbose=verbose,
        scheduler=fpx.scheduler(nosort=True),
    )


def sleepers_example_threads(verbose=True):
    """Example: scheduling is driven by number of threads."""
    sleepers_generic_program(
        verbose=verbose,
        scheduler=fpx.scheduler(limit='threads', max_threads=3),
    )


def sleepers_example_bindedthreads(verbose=True):
    """Example: scheduling is driven by number of threads and processes are binded."""
    sleepers_generic_program(
        verbose=verbose,
        scheduler=fpx.scheduler(limit='threads', max_threads=3, binded=True),
    )


def sleepers_example_memory(verbose=True):
    """Example: scheduling is driven by memory consumption."""
    sleepers_generic_program(
        verbose=verbose,
        scheduler=fpx.scheduler(limit='memory', memory_per_task=1.8),
    )


def sleepers_example_bindedmemory(verbose=True):
    """Example: scheduling is driven by memory consumption and processes are binded."""
    sleepers_generic_program(
        verbose=verbose,
        scheduler=fpx.scheduler(limit='memory', binded=True),
    )


def logarithmer_example(verbose=True):
    """Example: how to use a numpy array, shared among workers."""
    import numpy
    # sample initialization of the SharedNumpyArray
    nrows = 2
    ncols = 3
    a = SharedNumpyArray(numpy.ones((nrows, ncols)))
    for i in range(a.shape[0]):
        for j in range(a.shape[1]):
            a[i, j] = i + j
    print(a[...])
    # run workers, each one dedicated to a row
    boss = run_as_server(common_instructions={},
                         individual_instructions=dict(row=list(range(nrows))),
                         scheduler=fpx.scheduler(limit='threads', max_threads=nrows),
                         verbose=verbose,
                         sharedmemory_common_instructions=dict(array=a))
    boss.wait_till_finished()
    print(a[...])


def summer_example(verbose=True, use_lock=True):
    """Example: how to use a process-safe numpy array, shared among workers."""
    import numpy
    # sample initialization of the SharedNumpyArray
    n = 10
    s = SharedNumpyArray(numpy.zeros((1,), dtype=int))
    # run workers, each one dedicated to a value
    boss = run_as_server(common_instructions=dict(use_lock=use_lock),
                         individual_instructions=dict(value=list(range(n))),
                         # maximize the number of simultaneous threads to test the importance of the lock
                         scheduler=fpx.scheduler(limit='threads', max_threads=n),
                         verbose=verbose,
                         sharedmemory_common_instructions=dict(shared_sum=s))
    boss.wait_till_finished()
    print(str(s[0]) + '==' + str(sum(range(n))))


def matrixproduct_example(A_shape=(6, 7), B_shape=(7, 8),
                          iblocks=2, jblocks=4,
                          verbose=True):
    """
    Example: how to use several numpy array shared among workers.

    :param A_shape: size of the A matrix
    :param B_shape: size of the B matrix
    :param iblocks: number of blocks in direction i of result matrix;
                    must divide A_shape[0]
    :param jblocks: number of blocks in direction j of result matrix;
                    must divide B_shape[1]
    """
    import numpy
    assert A_shape[1] == B_shape[0]
    assert A_shape[0] % iblocks == 0 and B_shape[1] % jblocks == 0
    iblocksize = A_shape[0] // iblocks
    jblocksize = B_shape[1] // jblocks
    # sample initialization of the SharedNumpyArray
    A = SharedNumpyArray(numpy.random.random(A_shape))
    B = SharedNumpyArray(numpy.random.random(B_shape))
    C = SharedNumpyArray(numpy.zeros((A.shape[0], B.shape[1])))
    # build blocks indices, other cuts could be possible of course...
    i1 = list(range(0, C.shape[0], iblocksize))
    i2 = list(i1[1:] + [C.shape[0] - 1])
    j1 = list(range(0, C.shape[1], jblocksize))
    j2 = list(j1[1:] + [C.shape[1] - 1])
    t0 = time.time()
    # constitute each workers instructions
    indexes = dict(i1=[], i2=[], j1=[], j2=[])
    for i in range(len(i1)):
        for j in range(len(j1)):
            indexes['i1'].append(i1[i])
            indexes['i2'].append(i2[i])
            indexes['j1'].append(j1[j])
            indexes['j2'].append(j2[j])
    # run workers, each one dedicated to a block
    boss = run_as_server(common_instructions=dict(),
                         individual_instructions=indexes,
                         scheduler=fpx.scheduler(limit='threads', max_threads=4),
                         verbose=verbose,
                         sharedmemory_common_instructions=dict(A=A, B=B, C=C))
    boss.wait_till_finished()
    print('Exec in:', time.time() - t0)
    print(C[:, :])
